import torch
from torch.utils.data import Dataset

import psdenoiser.sde as sde_lib


def get_model_fn(model, train=False):
    """Create a function to give the output of the score-based model.

    Args:
      model: The score model.
      train: `True` for training and `False` for evaluation.

    Returns
    -------
      A model function.
    """

    def model_fn(x, t, x_cdn=None, cdn=None):
        """Compute the output of the score-based model.

        Args:
          x: A mini-batch of input data.
          labels: A mini-batch of conditioning variables for time steps. Should be interpreted differently
            for different models.

        Returns
        -------
          A tuple of (model output, new mutable states)
        """
        if not train:
            model.eval()
            return model(x, t, x_cdn=x_cdn, cdn=cdn)
        model.train()
        return model(x, t, x_cdn=x_cdn, cdn=cdn)

    return model_fn


def get_score_fn(sde, model, train=False, continuous=False):
    """Wraps `score_fn` so that the model output corresponds to a real time-dependent score function.

    Args:
      sde: An `sde_lib.SDE` object that represents the forward SDE.
      model: A score model.
      train: `True` for training and `False` for evaluation.
      continuous: If `True`, the score-based model is expected to directly take continuous time steps.

    Returns
    -------
      A score function.
    """
    model_fn = get_model_fn(model, train=train)

    if isinstance(sde, sde_lib.VPSDE) or isinstance(sde, sde_lib.subVPSDE):

        def score_fn(x, t, x_cdn=None, cdn=None):
            # Scale neural network output by standard deviation and flip sign
            if continuous or isinstance(sde, sde_lib.subVPSDE):
                # For VP-trained models, t=0 corresponds to the lowest noise level
                # The maximum value of time embedding is assumed to 999 for
                # continuously-trained models.
                time = t * 999
                score = model_fn(x, time, x_cdn=x_cdn, cdn=cdn)
                std = sde.marginal_prob(torch.zeros_like(x), t)[1]
            else:
                # For VP-trained models, t=0 corresponds to the lowest noise level
                time = t * (sde.N - 1)
                score = model_fn(x, time, x_cdn=x_cdn, cdn=cdn)
                std = sde.sqrt_one_minus_alphas_cumprod.to(labels.device)[labels.long()]

            score = -score / std[:, None, None, None]
            return score

    elif isinstance(sde, sde_lib.VESDE):

        def score_fn(x, t, x_cdn=None, cdn=None):
            if continuous:
                time = sde.marginal_prob(torch.zeros_like(x), t)[1]
            else:
                # For VE-trained models, t=0 corresponds to the highest noise level
                time = sde.T - t
                time *= sde.N - 1
                time = torch.round(time).long()

            score = model_fn(x, time, x_cdn=x_cdn, cdn=cdn)
            return score

    else:
        raise NotImplementedError(
            f"SDE class {sde.__class__.__name__} not yet supported."
        )

    return score_fn


class PS_Dataset(Dataset):
    def __init__(self, features, labels, cond):
        self.labels = labels
        self.features = features
        self.cond = cond

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        image = self.features[idx]
        label = self.labels[idx]
        cond = self.cond[idx]
        return image, label, cond



"""Based on Yang Song's code score_sde_pytorch/blob/main/losses.py ."""

from collections.abc import Callable

import torch

import psdenoiser.model_utils as mutils
from psdenoiser.sde import SDE
from psdenoiser.utils import extract

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


def forward_diffusion(
    sde: SDE, x_0: torch.Tensor, t: torch.Tensor, noise: torch.Tensor | None = None
) -> torch.Tensor:
    """Forward diffusion (using Sohl-Dickstein+15)."""
    if noise is None:
        noise = torch.randn_like(x_0)

    sqrt_alphas_cumprod_t = extract(sde.sqrt_alphas_cumprod, t, x_0.shape)
    sqrt_one_minus_alphas_cumprod_t = extract(
        sde.sqrt_one_minus_alphas_cumprod, t, x_0.shape
    )

    return sqrt_alphas_cumprod_t * x_0 + sqrt_one_minus_alphas_cumprod_t * noise


def get_sde_loss_fnc(
    sde: SDE,
    train: bool,
    reduce_mean: bool = True,
    continuous: bool = True,
    likelihood_weighting: bool = False,
    eps: float = 1e-5,
) -> Callable:
    r"""Create a loss function for training with arbirary SDEs.

    Parameters
    ----------
    sde: sde.SDE
        An `sde.SDE` object that represents the forward SDE.
    train: bool
        `True` for training loss and `False` for evaluation loss.
    reduce_mean:
        If `True`, average the loss across data dimensions.
        Otherwise sum the loss across data dimensions.
    continuous:
        `True` indicates that the model is defined to take
        continuous time steps. Otherwise it requires
        ad-hoc interpolation to take continuous time steps.
    likelihood_weighting:
        If `True`, weigh the mixture of score matching losses
        according to https://arxiv.org/abs/2101.09258; otherwise use the weighting
        recommended in Song+20.
    eps: float, optional
        The smallest time step from which to sample.

    Returns
    -------
        A loss function.
    """
    reduce_op = (
        torch.mean
        if reduce_mean
        else lambda *args, **kwargs: 0.5 * torch.sum(*args, **kwargs)
    )

    def loss_fn(
        model: torch.nn.Module,
        batch: torch.Tensor,
        cdn: torch.Tensor | None = None,
        x_cdn: torch.Tensor | None = None,
    ):
        """Compute the loss function.

        Parameters
        ----------
        model: torch.nn.Module
            A score model.
        batch: torch.Tensor
            A mini-batch of training data.
        cdn: torch.Tensor, optional
            A mini-batch of conditioning variables
            for the fully-connected network (similar to diffusion time).
        x_cdn: torch.Tensor, optional
            A mini-batch of conditioning variables
            for the U-Net (i.e. 2D image data).

        Returns
        -------
            loss: Scalar that represents the average loss value across the mini-batch.
        """
        score_fn = mutils.get_score_fn(sde, model, train=train, continuous=continuous)
        t = torch.rand(batch.shape[0], device=batch.device) * (sde.T - eps) + eps
        z = torch.randn_like(batch)
        # Calculate mean and std of p_t(x)
        mean, std = sde.marginal_prob(batch, t)
        perturbed_data = mean + std[:, None, None, None] * z
        # this returns -score/std
        score = score_fn(perturbed_data, t, x_cdn=x_cdn, cdn=cdn)

        if not likelihood_weighting:
            # this cancels the std, so this is -score + z
            losses = torch.square(score * std[:, None, None, None] + z)
            losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1)
        else:
            # This keeps the std: -score/std + z/std
            g2 = sde.sde(torch.zeros_like(batch), t)[1] ** 2
            losses = torch.square(score + z / std[:, None, None, None])
            # Then weighted by diffusion coefficient squared
            losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1) * g2

        return torch.mean(losses)

    return loss_fn



import torch


def extract(a: torch.Tensor, t: torch.Tensor, x_shape: tuple) -> torch.Tensor:
    r"""Extract the value of a tensor `a` at time `t` for a batch with shape `x_shape`.

    Parameters
    ----------
    a : torch.Tensor
        The tensor from which to extract values.
    t : torch.Tensor
        The tensor containing the time steps at which to extract values.
    x_shape : tuple
        The shape of the data for which to extract values.

    Returns
    -------
      torch.Tensor
        A tensor containing the extracted values,
        reshaped to match the batch size of `t`.
    """
    batch_size = t.shape[0]
    out = a.cpu().gather(-1, t.cpu())
    return out.reshape(batch_size, *((1,) * (len(x_shape) - 1))).to(t.device)


def transform(x: torch.Tensor, scale: torch.Tensor, bias: torch.Tensor) -> torch.Tensor:
    r"""Transform input x to be in [-1, 1] by first taking log10.

    Parameters
    ----------
    x : torch.Tensor
        The input tensor to be transformed.
    scale : torch.Tensor
        The scale factor to be applied.
    bias : torch.Tensor
        The bias to be applied.

    Returns
    -------
    torch.Tensor
        The transformed tensor in the range [-1, 1].
    """
    d = torch.log10(x)
    unit = (d - bias) / scale
    return unit * 2 - 1  # scale to [-1, 1]


def reverse_transform(
    y: torch.Tensor, scale: torch.Tensor, bias: torch.Tensor
) -> torch.Tensor:
    r"""Reverse the transformation from [-1, 1] to the original scale.

    Parameters
    ----------
    y : torch.Tensor
        The input tensor in the range [-1, 1].
    scale : torch.Tensor
        The scale factor used in the original transformation.
    bias : torch.Tensor
        The bias used in the original transformation.

    Returns
    -------
    torch.Tensor
        The tensor transformed back to the original scale.
    """
    unit = (y + 1) / 2
    d = unit * scale + bias
    return 10**d


def to_flattened_numpy(x):
    """Flatten a torch tensor `x` and convert it to numpy."""
    return x.detach().cpu().numpy().reshape((-1,))


def from_flattened_numpy(x, shape):
    """Form a torch tensor with the given `shape` from a flattened numpy array `x`."""
    return torch.from_numpy(x.reshape(shape))



"""Module that interacts with the Denoiser PyTorch model."""

from __future__ import annotations

import gc
import logging
import warnings
from pathlib import Path
from typing import Any

import numpy as np
import torch
from astropy import units as un
from tqdm.auto import tqdm

from psdenoiser import utils
from psdenoiser.inputs import DenoiserInput
from psdenoiser.model import UNet
from psdenoiser.outputs import DenoiserOutput
from psdenoiser.properties import denoiser_csts
from psdenoiser.sample_pytorch import GetODESampler
from psdenoiser.sde import VPSDE

log = logging.getLogger(__name__)


class Denoiser:
    r"""A class that loads 21cmPSDenoiser model and runs it on the input PS samples.

    Parameters
    ----------
    device : torch.device, optional
        The device on which to store the model.
        Default will use GPU is available, otherwise CPU.
    nsamples : int, optional
        Number of diffusion samples, default is 200.
    sampler_denoise : bool, optional
        If `True`, add one-step denoising to final samples, default is `True`.
    sampler_rtol : float, optional
        The relative tolerance level of the probability flow ODE solver,
        default is 1e-5.
    sampler_atol : float, optional
        The absolute tolerance level of the probability flow ODE solver,
        default is 1e-5.
    """

    def __init__(
        self,
        device: torch.device | None = None,
        nsamples: int = 200,
        sampler_denoise: bool = True,
        sampler_rtol: float = 1e-5,
        sampler_atol: float = 1e-5,
    ):
        if device is None:
            device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )
        self.csts = denoiser_csts
        model = UNet(
            dim=(len(self.csts.kperp), len(self.csts.kpar)),
            channels=2,
            dim_mults=(
                1,
                2,
                4,
                8,
            ),
            cdn_len=None,
        )
        here = Path(__file__).parent
        model.load_state_dict(
            torch.load(here / "denoiser_model.pt", map_location=device)
        )
        model.to(device)
        model.eval()
        self.model = model

        self.nsamples = nsamples
        sde = VPSDE(beta_min=0.1, beta_max=20.0)  # Like Ho+20
        self.sample = GetODESampler(
            sde,
            (nsamples, 1, len(self.csts.kperp), len(self.csts.kpar)),
            device=device,
            denoise=sampler_denoise,
            rtol=sampler_rtol,
            atol=sampler_atol,
        ).get_ode_sampler()

    def __getattr__(self, name: str) -> Any:
        """Allow access to denoiser properties directly from the denoiser object."""
        return getattr(self.csts, name)

    @torch.no_grad()
    def get_pred_single(self, noisy_sample: torch.Tensor) -> np.ndarray:
        r"""Get the mean 21-cm PS for a single noisy PS sample.

        Parameters
        ----------
        noisy_sample : torch.Tensor
            A single noisy PS sample of shape [len(kperp), len(kpar)].

        Returns
        -------
        samples_w_units : np.ndarray
            The denoised PS sample in mK^2,
            with shape [len(kperp), len(kpar)].
        """
        samples = (
            self.sample(self.model, x_cdn=noisy_sample[np.newaxis, ...], progress=True)
            .squeeze()
            .cpu()
            .detach()
        )

        samples_w_units = (
            utils.reverse_transform(samples, self.csts.mean_scale, self.csts.mean_bias)
            .cpu()
            .detach()
            .numpy()
        )

        noisy_sample.cpu()
        del noisy_sample
        return samples_w_units

    @torch.no_grad()
    def get_pred(
        self, noisy_samples: np.ndarray, progress: bool = True
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""Get the mean 21-cm PS for multiple noisy PS realisations.

        Parameters
        ----------
        noisy_samples : np.ndarray
            Noisy 21-cm PS realisations
            of shape [Nrealisations, len(kperp), len(kpar)] in mK^2.
        progress : bool, optional
            Whether to show a progress bar.
            Useful when passing a large batch of realisations.

        Returns
        -------
        all_samples : np.ndarray
            Full output with all denoised PS samples,
            shape [Nrealisations, Nsamples, len(kperp), len(kpar)].
        all_preds : np.ndarray
            The median of the denoiser PS
            taken over the diffusion samples `Nsamples`,
            shape [Nrealisations, len(kperp), len(kpar)].
        all_stds : np.ndarray
            The standard deviation of the denoised PS samples
            taken over the diffusion samples`Nsamples`,
            shape [Nrealisations, len(kperp), len(kpar)].
        """
        all_preds = []
        all_stds = []
        all_samples = []
        if progress:
            pbar = tqdm(
                range(noisy_samples.shape[0]),
                total=noisy_samples.shape[0],
                desc="Sampling ",
            )
        else:
            pbar = range(noisy_samples.shape[0])
        for i in pbar:
            samples = self.get_pred_single(torch.Tensor(noisy_samples[i]))
            all_preds.append(np.median(samples, axis=0))
            all_stds.append(np.std(samples, axis=0))
            all_samples.append(samples)
        return np.array(all_samples), np.array(all_preds), np.array(all_stds)

    def predict(
        self,
        ps_realisations: un.Quantity,
        kperp: un.Quantity,
        kpar: un.Quantity,
    ) -> DenoiserOutput:
        r"""Call 21cmPSDenoiser and predict the mean 21-cm PS for the given PS samples.

        Parameters
        ----------
        ps_realisations : un.Quantity
            cylindrical 21-cm PS in mK^2 of shape [N, len(kperp), len(kpar)]
            No NaNs or Infs allowed.
            If the mean of the PS realisation is < 1e-2mK^2,
            the denoiser will not be applied.
        kperp : un.Quantity
            kperp bin center values of the cylindrical PS
        kpar : un.Quantity
            kpar bin center values of the cylindrical PS
        N : int, optional
            Number of diffusion samples to take the median over to obtain the
            denoised result, default is 250.


        Returns
        -------
        DenoiserOutput
            See the class definition for more information.
        """
        if len(ps_realisations.shape) == 2:
            ps_realisations = ps_realisations[np.newaxis, ...]

        mask = np.mean(ps_realisations.value, axis=(-1, -2)) > self.csts.min_PS_mean

        if np.sum(mask) > 0:
            normed_ps_realisations, kperp, kpar = DenoiserInput().format_input(
                ps_realisations[mask], kperp, kpar
            )
            if np.sum(np.isnan(normed_ps_realisations)) > 0:
                raise ValueError("There are NaNs in the normalised input PS!!")
            samples_pred, med_pred, std_pred = self.get_pred(normed_ps_realisations)
            if np.sum(mask) < len(mask):
                warnings.warn(
                    f"Mean of PS is too low, skipping denoising"
                    f" for {len(mask) - np.sum(mask)} samples...",
                    stacklevel=2,
                )
                final_med = np.zeros_like(ps_realisations)
                final_std = np.zeros_like(ps_realisations)
                final_samples = np.zeros(
                    (
                        len(mask),
                        self.Nsamples,
                    )
                    + ps_realisations.shape[1:]
                )
                final_samples[mask] = samples_pred
                final_samples[~mask] = ps_realisations[:, None, ...][~mask]
                final_med[mask] = med_pred
                final_med[~mask] = ps_realisations[~mask]
                final_std[mask] = std_pred
                final_std[~mask] = np.nan
            else:
                final_med = med_pred
                final_std = std_pred
                final_samples = samples_pred
        else:
            final_med = ps_realisations
            final_std = np.ones_like(ps_realisations) + np.nan
            final_samples = ps_realisations
        gc.collect()
        return DenoiserOutput(
            final_samples.squeeze() * un.mK**2,
            final_med.squeeze() * un.mK**2,
            final_std.squeeze() * un.mK**2,
            kperp / un.Mpc,
            kpar / un.Mpc,
        )



"""Based on Yang Song's score_sde_pytorch/blob/main/sde_lib.py ."""

import abc

import numpy as np
import torch


class SDE(abc.ABC):
    r"""SDE abstract class. Functions are designed for a mini-batch of inputs."""

    def __init__(self, n):
        r"""Construct an SDE.

        Parameters
        ----------
          N: number of discretization time steps.
        """
        super().__init__()
        self.N = n

    @property
    @abc.abstractmethod
    def T(self):
        r"""End time of the SDE."""

    @abc.abstractmethod
    def sde(self, x, t):
        r"""Drift and diffusion functions for the forward SDE."""

    @abc.abstractmethod
    def marginal_prob(self, x, t):
        r"""Parameters to determine the marginal distribution of the SDE, $p_t(x)$."""

    @abc.abstractmethod
    def prior_sampling(self, shape):
        r"""Generate one sample from the prior distribution, $p_T(x)$."""

    @abc.abstractmethod
    def prior_logp(self, z):
        r"""Compute log-density of the prior distribution.

        Useful for computing the log-likelihood via probability flow ODE.

        Parameters
        ----------
          z: latent code

        Returns
        -------
            torch.Tensor
                Log probability density of the prior distribution.
                The shape is (batch_size,).
        """

    def discretize(self, x, t):
        r"""Discretize the SDE in the form: x_{i+1} = x_i + f_i(x_i) + G_i z_i.

        Useful for reverse diffusion sampling and probabiliy flow sampling.
        Defaults to Euler-Maruyama discretization.

        Args:
            x: a torch tensor
            t: a torch float representing the time step (from 0 to `self.T`)

        Returns
        -------
            f, G
        """
        dt = 1 / self.N
        drift, diffusion = self.sde(x, t)
        f = drift * dt
        G = diffusion * torch.sqrt(torch.tensor(dt, device=t.device))
        return f, G

    def reverse(self, score_fn: torch.Tensor, probability_flow: bool = False):
        r"""Create the reverse-time SDE/ODE.

        Parameters
        ----------
            score_fn:
                A time-dependent score-based model that
                takes x and t and returns the score.
            probability_flow:
                If `True`, create the reverse-time ODE
                used for probability flow sampling.
        """
        N = self.N
        T = self.T
        sde_fn = self.sde
        discretize_fn = self.discretize

        class RSDE(self.__class__):
            def __init__(self):
                r"""Class for the reverse-time SDE."""
                self.N = N
                self.probability_flow = probability_flow

            @property
            def T(self):
                r"""End time of the SDE."""
                return T

            def sde(
                self,
                x: torch.Tensor,
                t: torch.Tensor,
                x_cdn: torch.Tensor | None = None,
                cdn: torch.Tensor | None = None,
            ) -> tuple[torch.Tensor, torch.Tensor]:
                r"""Drift and diffusion functions for the reverse SDE/ODE."""
                drift, diffusion = sde_fn(x, t)
                score = score_fn(x, t, x_cdn=x_cdn, cdn=cdn)
                drift = drift - diffusion[:, None, None, None] ** 2 * score * (
                    0.5 if self.probability_flow else 1.0
                )
                # Set the diffusion function to zero for ODEs.
                diffusion = 0.0 if self.probability_flow else diffusion
                return drift, diffusion

            def discretize(
                self,
                x: torch.Tensor,
                t: torch.Tensor,
                x_cdn: torch.Tensor | None = None,
                cdn: torch.Tensor | None = None,
            ):
                r"""Discretized iteration rules for the reverse diffusion sampler."""
                f, G = discretize_fn(x, t)
                rev_f = f - G[:, None, None, None] ** 2 * score_fn(
                    x, t, x_cdn=x_cdn, cdn=cdn
                ) * (0.5 if self.probability_flow else 1.0)
                rev_G = torch.zeros_like(G) if self.probability_flow else G
                return rev_f, rev_G

        return RSDE()


class VPSDE(SDE):
    def __init__(self, beta_min: float = 0.1, beta_max: float = 20, n: float = 1000):
        r"""Construct a Variance Preserving SDE.

        Args:
          beta_min: value of beta(0)
          beta_max: value of beta(1)
          n: number of discretization steps
        """
        super().__init__(n)
        self.beta_0 = beta_min
        self.beta_1 = beta_max
        self.N = n
        self.discrete_betas = torch.linspace(beta_min / n, beta_max / n, n)
        self.alphas = 1.0 - self.discrete_betas
        self.alphas_cumprod = torch.cumprod(self.alphas, dim=0)
        self.sqrt_alphas_cumprod = torch.sqrt(self.alphas_cumprod)
        self.sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - self.alphas_cumprod)

    @property
    def T(self):
        r"""End time of the SDE."""
        return 1

    def sde(self, x, t):
        r"""Drift and diffusion functions for the forward SDE."""
        beta_t = self.beta_0 + t * (self.beta_1 - self.beta_0)
        drift = -0.5 * beta_t[:, None, None, None] * x
        diffusion = torch.sqrt(beta_t)
        return drift, diffusion

    def marginal_prob(self, x, t):
        r"""Parameters to determine the marginal distribution of the SDE, $p_t(x)$."""
        log_mean_coeff = (
            -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        )
        mean = torch.exp(log_mean_coeff[:, None, None, None]) * x
        std = torch.sqrt(1.0 - torch.exp(2.0 * log_mean_coeff))
        return mean, std

    def prior_sampling(self, shape):
        r"""Generate one sample from the prior distribution, $p_T(x)$."""
        return torch.randn(*shape)

    def prior_logp(self, z):
        r"""Compute log-density of the prior distribution."""
        shape = z.shape
        N = np.prod(shape[1:])
        return -N / 2.0 * np.log(2 * np.pi) - torch.sum(z**2, dim=(1, 2, 3)) / 2.0

    def discretize(self, x, t):
        """DDPM discretization."""
        timestep = (t * (self.N - 1) / self.T).long()
        beta = self.discrete_betas.to(x.device)[timestep]
        alpha = self.alphas.to(x.device)[timestep]
        sqrt_beta = torch.sqrt(beta)
        f = torch.sqrt(alpha)[:, None, None, None] * x - x
        G = sqrt_beta
        return f, G


class SubVPSDE(SDE):
    def __init__(self, beta_min: float = 0.1, beta_max: float = 20, n: int = 1000):
        r"""Construct the sub-VP SDE that excels at likelihoods.

        Parameters
        ----------
            beta_min: float
                Value of beta(0).
            beta_max: float
                Value of beta(1).
            n: int
                Number of discretization steps.
        """
        super().__init__(n)
        self.beta_0 = beta_min
        self.beta_1 = beta_max
        self.N = n

    @property
    def T(self):
        r"""End time of the SDE."""
        return 1

    def sde(self, x, t):
        r"""Drift and diffusion functions for the forward SDE."""
        beta_t = self.beta_0 + t * (self.beta_1 - self.beta_0)
        drift = -0.5 * beta_t[:, None, None, None] * x
        discount = 1.0 - torch.exp(
            -2 * self.beta_0 * t - (self.beta_1 - self.beta_0) * t**2
        )
        diffusion = torch.sqrt(beta_t * discount)
        return drift, diffusion

    def marginal_prob(self, x, t):
        r"""Parameters to determine the marginal distribution of the SDE, $p_t(x)$."""
        log_mean_coeff = (
            -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        )
        mean = torch.exp(log_mean_coeff)[:, None, None, None] * x
        std = 1 - torch.exp(2.0 * log_mean_coeff)
        return mean, std

    def prior_sampling(self, shape):
        r"""Generate one sample from the prior distribution, $p_T(x)$."""
        return torch.randn(*shape)

    def prior_logp(self, z):
        r"""Compute log-density of the prior distribution."""
        shape = z.shape
        N = np.prod(shape[1:])
        return -N / 2.0 * np.log(2 * np.pi) - torch.sum(z**2, dim=(1, 2, 3)) / 2.0


class VESDE(SDE):
    def __init__(self, sigma_min: float = 0.01, sigma_max: float = 50, n: float = 1000):
        r"""Construct a Variance Exploding SDE.

        Parameters
        ----------
            sigma_min: float
                Smallest sigma, default is 0.01.
            sigma_max: float
                Largest sigma, default is 50.
            n: int
                Number of discretization steps, default is 1000.
        """
        super().__init__(n)
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.discrete_sigmas = torch.exp(
            torch.linspace(np.log(self.sigma_min), np.log(self.sigma_max), n)
        )
        self.N = n

    @property
    def T(self):
        r"""End time of the SDE."""
        return 1

    def sde(self, x, t):
        r"""Drift and diffusion functions for the forward SDE."""
        sigma = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        drift = torch.zeros_like(x)
        diffusion = sigma * torch.sqrt(
            torch.tensor(
                2 * (np.log(self.sigma_max) - np.log(self.sigma_min)), device=t.device
            )
        )
        return drift, diffusion

    def marginal_prob(self, x, t):
        r"""Parameters to determine the marginal distribution of the SDE, $p_t(x)$."""
        std = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        mean = x
        return mean, std

    def prior_sampling(self, shape):
        r"""Generate one sample from the prior distribution, $p_T(x)$."""
        return torch.randn(*shape) * self.sigma_max

    def prior_logp(self, z):
        r"""Compute log-density of the prior distribution."""
        shape = z.shape
        N = np.prod(shape[1:])
        return -N / 2.0 * np.log(2 * np.pi * self.sigma_max**2) - torch.sum(
            z**2, dim=(1, 2, 3)
        ) / (2 * self.sigma_max**2)

    def discretize(self, x, t):
        """SMLD(NCSN) discretization."""
        timestep = (t * (self.N - 1) / self.T).long()
        sigma = self.discrete_sigmas.to(t.device)[timestep]
        adjacent_sigma = torch.where(
            timestep == 0,
            torch.zeros_like(t),
            self.discrete_sigmas[timestep - 1].to(t.device),
        )
        f = torch.zeros_like(x)
        G = torch.sqrt(sigma**2 - adjacent_sigma**2)
        return f, G



"""Module to organise the denoiser output."""

from __future__ import annotations

import dataclasses as dc
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path

import astropy.units as un
import numpy as np

from psdenoiser.properties import denoiser_csts


@dataclass(frozen=True)
class DenoiserOutput:
    """A simple class that makes it easier to access the denoiser output."""

    deltasq_samples: un.Quantity
    deltasq_median: un.Quantity
    deltasq_std: un.Quantity
    kperp: un.Quantity
    kpar: un.Quantity

    csts = denoiser_csts

    def keys(self) -> Generator[str, None, None]:
        """Yield the keys of the main data products."""
        for k in dc.fields(self):
            yield k.name

    def items(self) -> Generator[tuple[str, np.ndarray], None, None]:
        """Yield the keys and values of the main data products, like a dict."""
        for k in self.keys():
            yield k, getattr(self, k)

    def __getitem__(self, key: str) -> np.ndarray:
        """Allow access to attributes as items."""
        return getattr(self, key)

    def squeeze(self):
        """Return a new EmulatorOutput with all dimensions of length 1 removed."""
        return DenoiserOutput(**{k: np.squeeze(v) for k, v in self.items()})

    @property
    def denoiser_median_on_test_mean(self):
        """Return the denoiser median error on the test set."""
        return (
            self.csts.denoiser_median_on_test_mean_percent / 100.0 * self.deltasq_median
        )

    def write(
        self,
        fname: str | Path,
        noisy_ps: np.ndarray | None = None,
        store: list[str] | None = None,
        clobber: bool = False,
    ):
        """Write this instance's data to a file.

        This saves the output as a numpy .npz file. The output is saved as a dictionary
        with the keys being the names of the attributes of this class and the values
        being the corresponding values of those attributes. If theta is not None, then
        the inputs are also saved under the key "inputs".

        Parameters
        ----------
        fname : str or Path
            The filename to write to.
        noisy_ps : np.ndarray or dict or None, optional
            The input noisy cylindrical PS associated with this output data to write to the file.
            If None, the inputs are not written.
        store : list of str or None, optional
            The names of the attributes to write to the file. If None, all attributes
            are written.
        clobber : bool, optional
            Whether to overwrite the file if it already exists.
        """
        if store is None:
            store = list(self.__dict__.keys())

        pth = Path(fname)
        if pth.exists() and not clobber:
            raise ValueError(f"File {pth} exists and clobber=False.")

        out = {k: getattr(self, k) for k in store}
        if noisy_ps is not None:
            out["inputs"] = noisy_ps

        np.savez(fname, out)



from denoiser import Denoiser
import numpy as np

path = '/projects/cosmo_database/dbreitman/CV_PS/Full_May2023_DB/'
f = np.load(path + 'dec_db_50_thetas_nointerp_nolog_lesszs.npz')
seeds = f['PS_2D_seeds']
means = f['PS_2D_means']
kperp = f['kperp']
kpar = f['kpar']
std_means = f['PS_2D_std_means']

shape = (means.shape[0] * means.shape[1], means.shape[2], means.shape[3])
means = means.reshape(shape)
std_means = std_means.reshape(shape)
normed_in, out = Denoiser().predict(seeds.reshape(shape), kperp, kpar)
print('MEAN', np.mean(abs((out.mean_PS - means) / means)))
print('STD', np.mean(abs((out.std_PS - std_means) / std_means)))
print(out.mean_PS.shape, std_mean.shape)


"""Based on "The Annotated Diffusion Model" from Hugging Face."""
import math
from inspect import isfunction
from functools import partial

from einops import rearrange, reduce
from einops.layers.torch import Rearrange

import torch
from torch import nn, einsum
import torch.nn.functional as F

# use seed for reproducability
torch.manual_seed(0)

def exists(x):
    return x is not None

def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d

class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x


def Upsample(dim, dim_out=None):
    return nn.Sequential(
        nn.Upsample(scale_factor=2, mode="nearest"),
        nn.Conv2d(dim, default(dim_out, dim), 3, padding=1),
    )


def Downsample(dim, dim_out=None):
    # No More Strided Convolutions or Pooling
    return nn.Sequential(
        Rearrange("b c (h p1) (w p2) -> b (c p1 p2) h w", p1=2, p2=2),
        nn.Conv2d(dim * 4, default(dim_out, dim), 1),
    )


class SinusoidalPositionEmbeddings(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, time):
        device = time.device
        half_dim = self.dim // 2
        embeddings = math.log(10000) / (half_dim - 1)
        embeddings = torch.exp(torch.arange(half_dim, device=device) * -embeddings)
        embeddings = time[:, None] * embeddings[None, :]
        embeddings = torch.cat((embeddings.sin(), embeddings.cos()), dim=-1)
        return embeddings


class WeightStandardizedConv2d(nn.Conv2d):
    """
    https://arxiv.org/abs/1903.10520
    weight standardization purportedly works synergistically with group normalization
    """

    def forward(self, x):
        eps = 1e-5 if x.dtype == torch.float32 else 1e-3

        weight = self.weight
        mean = reduce(weight, "o ... -> o 1 1 1", "mean")
        var = reduce(weight, "o ... -> o 1 1 1", partial(torch.var, unbiased=False))
        normalized_weight = (weight - mean) / (var + eps).rsqrt()

        return F.conv2d(
            x,
            normalized_weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )


class Block(nn.Module):
    def __init__(self, dim, dim_out, kernel_size=3, padding=1, groups=8):
        super().__init__()
        self.proj = WeightStandardizedConv2d(dim, dim_out, kernel_size, padding=padding)
        self.norm = nn.GroupNorm(groups, dim_out)
        self.act = nn.SiLU()

    def forward(self, x, scale_shift=None):
        x = self.proj(x)
        x = self.norm(x)

        if exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1) + shift

        x = self.act(x)
        return x


class ResnetBlock(nn.Module):
    """https://arxiv.org/abs/1512.03385"""

    def __init__(self, dim, dim_out, *, kernel_size=3,time_emb_dim=None, groups=8):
        super().__init__()
        self.mlp = (
            nn.Sequential(nn.SiLU(), nn.Linear(time_emb_dim, dim_out * 2))
            if exists(time_emb_dim)
            else None
        )

        self.block1 = Block(dim, dim_out, kernel_size = kernel_size,groups=groups)
        self.block2 = Block(dim_out, dim_out, kernel_size = kernel_size,groups=groups)
        self.res_conv = nn.Conv2d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self, x, time_emb=None):
        scale_shift = None
        if exists(self.mlp) and exists(time_emb):
            time_emb = self.mlp(time_emb)
            time_emb = rearrange(time_emb, "b c -> b c 1 1")
            scale_shift = time_emb.chunk(2, dim=1)

        h = self.block1(x, scale_shift=scale_shift)
        h = self.block2(h)
        return h + self.res_conv(x)

class Attention(nn.Module):
    def __init__(self, dim, heads=4, dim_head=32):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        hidden_dim = dim_head * heads
        self.to_qkv = nn.Conv2d(dim, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Conv2d(hidden_dim, dim, 1)

    def forward(self, x):
        b, c, h, w = x.shape
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads), qkv
        )
        q = q * self.scale

        sim = einsum("b h d i, b h d j -> b h i j", q, k)
        sim = sim - sim.amax(dim=-1, keepdim=True).detach()
        attn = sim.softmax(dim=-1)

        out = einsum("b h i j, b h d j -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)
        return self.to_out(out)

class LinearAttention(nn.Module):
    def __init__(self, dim, heads=4, dim_head=32):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        hidden_dim = dim_head * heads
        self.to_qkv = nn.Conv2d(dim, hidden_dim * 3, 1, bias=False)

        self.to_out = nn.Sequential(nn.Conv2d(hidden_dim, dim, 1), 
                                    nn.GroupNorm(1, dim))

    def forward(self, x):
        b, c, h, w = x.shape
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads), qkv
        )

        q = q.softmax(dim=-2)
        k = k.softmax(dim=-1)

        q = q * self.scale
        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)

        out = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        out = rearrange(out, "b h c (x y) -> b (h c) x y", h=self.heads, x=h, y=w)
        return self.to_out(out)

class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.fn = fn
        self.norm = nn.GroupNorm(1, dim)

    def forward(self, x):
        x = self.norm(x)
        return self.fn(x)



class UNet(nn.Module):
    def __init__(
        self,
        dim,
        init_dim=None,
        out_dim=None,
        dim_mults=(1, 2, 4, 8),
        channels=2,
        resnet_block_groups=4,
        cdn_len=None,
    ):
        super().__init__()

        # determine dimensions
        self.channels = channels
        input_channels = channels

        init_dim = 16
        self.init_conv = nn.Conv2d(input_channels, init_dim, 1, padding=0) # changed to 1 and 0 from 7,3

        dims = [init_dim, *map(lambda m: dim[0] * m, dim_mults)]
        in_out = list(zip(dims[:-1], dims[1:]))

        resnet_block = partial(ResnetBlock, groups=resnet_block_groups)

        # time embeddings
        time_dim = 128

        self.time_mlp = nn.Sequential(
            SinusoidalPositionEmbeddings(dim[0]),
            nn.Linear(dim[0], time_dim),
            nn.GELU(),
            nn.Linear(time_dim, time_dim),
        )
        if cdn_len is not None:
            self.cdn_mlp = nn.Sequential(
                nn.Linear(cdn_len, 64),
                nn.GELU(),
                nn.Linear(64, 64),
            )
            time_dim += 64

        # layers
        self.downs = nn.ModuleList([])
        self.ups = nn.ModuleList([])
        num_resolutions = len(in_out)
        for ind, (dim_in, dim_out) in enumerate(in_out):
            is_last = ind >= (num_resolutions - 1)

            self.downs.append(
                nn.ModuleList(
                    [
                        resnet_block(dim_in, dim_in, time_emb_dim=time_dim),
                        resnet_block(dim_in, dim_in, time_emb_dim=time_dim),
                        Residual(PreNorm(dim_in, LinearAttention(dim_in))),
                        Downsample(dim_in, dim_out)
                        if not is_last
                        else nn.Conv2d(dim_in, dim_out, 3, padding=1),
                    ]
                )
            )

        mid_dim = dims[-1]
        self.mid_block1 = resnet_block(mid_dim, mid_dim, time_emb_dim=time_dim)
        self.mid_attn = Residual(PreNorm(mid_dim, Attention(mid_dim)))
        self.mid_block2 = resnet_block(mid_dim, mid_dim, time_emb_dim=time_dim)

        for ind, (dim_in, dim_out) in enumerate(reversed(in_out)):
            is_last = ind == (len(in_out) - 1)

            self.ups.append(
                nn.ModuleList(
                    [
                        resnet_block(dim_out + dim_in, dim_out, time_emb_dim=time_dim),
                        resnet_block(dim_out + dim_in, dim_out, time_emb_dim=time_dim),
                        Residual(PreNorm(dim_out, LinearAttention(dim_out))),
                        Upsample(dim_out, dim_in)
                        if not is_last
                        else nn.Conv2d(dim_out, dim_in, 3, padding=1),
                    ]
                )
            )

        self.out_dim = 1

        self.final_res_block = resnet_block(init_dim * 2, dim[0], time_emb_dim=time_dim)
        self.final_conv = nn.ConvTranspose2d(dim[0], self.out_dim, 1)

    def forward(self, x, time, x_cdn=None, cdn=None):
        if x_cdn is not None:
            x = torch.cat((x, x_cdn), dim=1)
        x = self.init_conv(x)
        r = x.clone()
        t = self.time_mlp(time)
        if cdn is not None:
            c = self.cdn_mlp(cdn)
            t = torch.cat((t,c), axis = -1)

        h = []
        for block1, block2, attn, downsample in self.downs:
            x = block1(x, t)
            h.append(x)

            x = block2(x, t)
            x = attn(x)
            h.append(x)

            x = downsample(x)
        x = self.mid_block1(x, t)
        x = self.mid_attn(x)
        x = self.mid_block2(x, t)
        for block1, block2, attn, upsample in self.ups:
            x = torch.cat((x, h.pop()), dim=1)
            x = block1(x, t)

            x = torch.cat((x, h.pop()), dim=1)
            x = block2(x, t)
            x = attn(x)

            x = upsample(x)
        x = torch.cat((x, r), dim=1)

        x = self.final_res_block(x, t)
        return self.final_conv(x)


"""21cmPSDenoiser package."""

__all__ = [
    "VESDE",
    "VPSDE",
    "Denoiser",
    "DenoiserInput",
    "DenoiserOutput",
    "GetODESampler",
    "SubVPSDE",
    "UNet",
    "__version__",
    "denoiser_csts",
    "get_score_fn",
    "model_utils",
    "reverse_transform",
    "transform",
]
from . import model_utils
from ._version import __version__
from .denoiser import Denoiser
from .inputs import DenoiserInput
from .model import UNet
from .model_utils import get_score_fn
from .outputs import DenoiserOutput
from .properties import denoiser_csts
from .sample_pytorch import GetODESampler
from .sde import VESDE, VPSDE, SubVPSDE
from .utils import reverse_transform, transform



from pathlib import Path

import numpy as np
from astropy import units as un


class DenoiserConstants:
    """A class that contains the constants of the denoiser."""

    def __init__(self):
        here = Path(__file__).parent
        with np.load(here / "constants.npz") as f:
            self.noisy_bias = f["noisy_mean"]
            self.noisy_scale = f["noisy_std"]
            self.mean_bias = f["mean_mean"]
            self.mean_scale = f["mean_std"]
            self.param_labels = f["param_labels"]
            self.modes = f["Nmodes"]
            self.kperp = f["kperp"] / un.Mpc
            self.kpar = f["kpar"] / un.Mpc
            self.min_PS_mean = 1e-2
            self.denoiser_median_on_test_mean_percent = f["median_err_on_mean_2D"]
            self.h = f["h"]


denoiser_csts = DenoiserConstants()



"""From Yang Song's score_sde_pytorch/sampling.py
ODE sampler for probability flow models.
"""

import abc
import gc

import torch
from scipy import integrate

from psdenoiser.model_utils import get_score_fn
from psdenoiser.utils import from_flattened_numpy, to_flattened_numpy


class Predictor(abc.ABC):
    """The abstract class for a predictor algorithm."""

    def __init__(self, sde, score_fn, probability_flow=False):
        super().__init__()
        self.sde = sde
        # Compute the reverse SDE/ODE
        self.rsde = sde.reverse(score_fn, probability_flow)
        self.score_fn = score_fn

    @abc.abstractmethod
    def update_fn(self, x, t, x_cdn=None, cdn=None):
        """One update of the predictor.

        Args:
          x: A PyTorch tensor representing the current state
          t: A Pytorch tensor representing the current time step.

        Returns
        -------
          x: A PyTorch tensor of the next state.
          x_mean: A PyTorch tensor. The next state without random noise. Useful for denoising.
        """


class ReverseDiffusionPredictor(Predictor):
    def __init__(self, sde, score_fn, probability_flow=False):
        super().__init__(sde, score_fn, probability_flow)

    def update_fn(self, x, t, x_cdn=None, cdn=None):
        f, G = self.rsde.discretize(x, t, x_cdn=x_cdn, cdn=cdn)
        z = torch.randn_like(x)
        x_mean = x - f
        x = x_mean + G[:, None, None, None] * z
        return x, x_mean


class GetODESampler:
    def __init__(
        self,
        sde,
        shape,
        inverse_scaler=None,
        denoise=False,
        rtol=1e-5,
        atol=1e-5,
        method="RK45",
        eps=1e-3,
        device="cuda",
    ):
        """Probability flow ODE sampler with the black-box ODE solver.

        Args:
            sde: An `sde_lib.SDE` object that represents the forward SDE.
            shape: A sequence of integers. The expected shape of a single sample.
            inverse_scaler: The inverse data normalizer.
            denoise: If `True`, add one-step denoising to final samples.
            rtol: A `float` number. The relative tolerance level of the ODE solver.
            atol: A `float` number. The absolute tolerance level of the ODE solver.
            method: A `str`. The algorithm used for the black-box ODE solver.
              See the documentation of `scipy.integrate.solve_ivp`.
            eps: A `float` number. The reverse-time SDE/ODE will be integrated to `eps` for numerical stability.
            device: PyTorch device.

        Returns
        -------
            A sampling function that returns samples and the number of function evaluations during sampling.
        """
        self.sde = sde
        self.shape = shape
        self.inverse_scaler = inverse_scaler
        self.denoise = denoise
        self.rtol = rtol
        self.atol = atol
        self.method = method
        self.eps = eps
        self.device = device

    def denoise_update_fn(self, model, x, x_cdn=None, cdn=None):
        score_fn = get_score_fn(self.sde, model, train=False, continuous=True)
        # Reverse diffusion predictor for denoising
        predictor_obj = ReverseDiffusionPredictor(
            self.sde, score_fn, probability_flow=False
        )
        vec_eps = torch.ones(x.shape[0], device=x.device) * self.eps
        _, x = predictor_obj.update_fn(x, vec_eps, x_cdn=x_cdn, cdn=cdn)
        return x

    def drift_fn(self, model, x, t, x_cdn=None, cdn=None):
        """Get the drift function of the reverse-time SDE."""
        score_fn = get_score_fn(self.sde, model, train=False, continuous=True)
        rsde = self.sde.reverse(score_fn, probability_flow=True)
        return rsde.sde(x, t, x_cdn=x_cdn, cdn=cdn)[0]

    def ode_sampler(
        self,
        model: torch.nn.Module,
        z: torch.Tensor | None = None,
        x_cdn: torch.Tensor | None = None,
        cdn: torch.Tensor | None = None,
        progress: bool = False,
        return_nfe: bool = False,
    ):
        r"""Solves the probability flow ODE with black-box ODE solver.

        Parameters
        ----------
        All inputs are on CPU except model that's on the GPU already.

        model: torch.nn.Module
            A score model.
        z: torch.Tensor, optional
            If provided, generate samples from latent `z`.

        Returns
        -------
          samples, number of function evaluations.
        """
        with torch.no_grad():
            # Initial sample
            if z is None:
                # If not provided, sample from the prior distibution of the SDE.
                x = self.sde.prior_sampling(self.shape).to(self.device)
            else:
                x = z.to(self.device)
            if x_cdn is not None:
                x_cdn = (
                    x_cdn.repeat((self.shape[0],) + (1,) * len(self.shape[1:]))
                    .reshape(self.shape)
                    .to(self.device)
                )

            def ode_func(t, x):
                x = (
                    from_flattened_numpy(x, self.shape)
                    .to(self.device)
                    .type(torch.float32)
                )
                vec_t = torch.ones(self.shape[0], device=x.device) * t
                drift = self.drift_fn(
                    model, x, vec_t, x_cdn=x_cdn, cdn=cdn if cdn is not None else None
                )
                x.cpu()
                vec_t.cpu()
                return to_flattened_numpy(drift)

            # Black-box ODE solver for the probability flow ODE
            solution = integrate.solve_ivp(
                ode_func,
                (self.sde.T, self.eps),
                to_flattened_numpy(x),
                rtol=self.rtol,
                atol=self.atol,
                method=self.method,
            )
            x.cpu()
            nfe = solution.nfev
            sln = (
                torch.tensor(solution.y[:, -1])
                .reshape(self.shape)
                .to(self.device)
                .type(torch.float32)
            )
            # Denoising = running one predictor step without adding noise
            if self.denoise:
                sln = self.denoise_update_fn(model, sln, x_cdn=x_cdn, cdn=cdn)
            if self.inverse_scaler is not None:
                sln = self.inverse_scaler(sln)

            if x_cdn is not None:
                x_cdn.cpu()
            sln.cpu()
            gc.collect()
            torch.cuda.empty_cache()
            if return_nfe:
                return sln, nfe
            return sln

    def get_ode_sampler(self):
        return self.ode_sampler



"""Module containing functionality for handling denoiser inputs."""

import warnings

import astropy.units as un
import numpy as np
import torch

from psdenoiser.properties import denoiser_csts as csts
from psdenoiser.utils import transform


class DenoiserInput:
    """Class for handling denoiser input."""

    def format_input(
        self,
        ps_realisations: un.Quantity,
        kperp: un.Quantity,
        kpar: un.Quantity,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Format the input 2D PS into a 3D numpy array with shape (Nsamples, Nkperp, nkpar).

        Parameters
        ----------
        ps_realisations : un.Quantity
            Cylindrical 21-cm power spectrum sample(s) P(kperp, kpar) to be denoised
            in units of mK^2.
        kperp : un.Quantity
            K perpendicular modes of the cylindrical power spectrum in units of Mpc^{-1}
            by default.
        kpar : un.Quantity
            K parallel modes of the cylindrical power spectrum in units of Mpc^{-1}
            by default.
        """
        if not (np.allclose(kperp, csts.kperp) and np.allclose(kpar, csts.kpar)):
            raise NotImplementedError(
                "Interpolation of input kperp and kpar bins is not supported yet."
            )

        if ps_realisations.shape[1] != len(kperp):
            raise ValueError(
                "You supplied the wrong kperp bins: %s kperp bins in PS vs %s kperp bins in kperp array supplied."
                % (ps_realisations.shape[1], len(kperp))
            )
        if ps_realisations.shape[2] != len(kpar):
            raise ValueError(
                "You supplied the wrong kpar bins: %s kpar bins in PS vs %s kpar bins in kpar array supplied."
                % (ps_realisations.shape[2], len(kpar))
            )
        if len(ps_realisations.shape) > 3:
            raise ValueError(
                "The shape of the input ps_realisations PS should be (Nsamples, Nkperp, Nkpar)."
            )
        if np.sum(np.isnan(ps_realisations)):
            raise ValueError("There should not be any NaNs in the PS array!")

        # Check that input kperp and kpar are the same as the ones of the autoencoder data
        # In the future, this could be replaced by an interpolation step.

        if not np.allclose(kperp, csts.kperp):
            warnings.warn(
                "Input kperp bins are not the same as the expected kperp bins! Proceeding anyways..."
            )
        if not np.allclose(kpar, csts.kpar):
            warnings.warn(
                "Input kpar bins are not the same as the expected kpar bins! Proceeding anyways..."
            )
        if ps_realisations.unit != un.mK**2:
            raise ValueError(
                "Input PS should be in mK^2, but is in %s." % ps_realisations.unit
            )
        if kperp.unit != un.Mpc ** (-1):
            raise ValueError(
                "Input kperp should be in Mpc^{-1}, but is in %s." % kperp.unit
            )
        if kpar.unit != un.Mpc ** (-1):
            raise ValueError(
                "Input kpar should be in Mpc^{-1}, but is in %s." % kpar.unit
            )

        return self.normalize_noisy(ps_realisations.value), kperp.value, kpar.value

    def normalize_noisy(self, noisy: np.ndarray) -> np.ndarray:
        """Normalize the input PS to be in [-1,1]."""
        return (
            transform(
                torch.Tensor(noisy),
                torch.Tensor(csts.noisy_scale),
                torch.Tensor(csts.noisy_bias),
            )
            .cpu()
            .detach()
            .numpy()
        )



"""Plotting routines for plots in Breitman+25a"""

from astropy import units as u
from astropy.cosmology.units import littleh
from astropy.cosmology import Planck15 as cosmo
from scipy.constants import c
import matplotlib.colors as colors
import numpy as np
speed_of_light = c * u.m / u.s
from matplotlib import rcParams
import matplotlib.pyplot as plt
from scipy.interpolate import RectBivariateSpline, interp1d

def freq2z(f):
    return 1420.4 / f  - 1.

def z2freq(z):
    return 1420.4 / (z + 1.)

def alpha(z):
    nu_21 = z2freq(z) *u.MHz
    speed_of_light = c * u.m / u.s
    return (speed_of_light * (1+z) / (nu_21 * cosmo.H(z))).to(u.Mpc / u.MHz)

def delay2kpar(tau, freq = None, z = None):
    if z is None:
        z = freq2z(freq.value)
    if freq is None:
        freq = z2freq(z)*u.MHz
    kpar = 2*np.pi * np.abs(tau) / alpha(z)
    return kpar.to(1/u.Mpc)

def kperp2baseline(kperp, freq = None, z = None):
    if z is None:
        z = freq2z(freq.value)
    if freq is None:
        freq = z2freq(z)*u.MHz
    baseline = np.abs(kperp * cosmo.comoving_distance(z) * speed_of_light / (2 * np.pi * freq))
    return baseline.to(u.m)

def horizon_limit(kperp, freq = None, z = None, buffer=300*u.ns):
    if z is None:
        z = freq2z(freq.value)
    if freq is None:
        freq = z2freq(z)*u.MHz
    baseline = kperp2baseline(kperp, freq, z)
    tau_wedge = baseline / speed_of_light # Eqn 8 Josh's paper
    return delay2kpar(tau_wedge + buffer, freq, z)

def get_hera_var_at_z(redshift):
    with np.load('HERA_mock_sensitivities_nosample.npz', allow_pickle=True) as f:
        freq = z2freq(redshift)
        bands = f['band_MHz']
        my_band = np.argmin(abs(bands - freq))
        print("HERA redshift", np.round(freq2z(bands[my_band]),1), "vs input redshift", np.round(redshift,1))
        this_band = str(int(bands[my_band]))+'MHz'
        kperp_centers = np.exp((np.log(f['kperp_edges'][1:]) + np.log(f['kperp_edges'][:-1])) / 2)
        kpar_centers = (f['kpar_edges'][1:] + f['kpar_edges'][:-1])/2
        return freq2z(bands[my_band]), kperp_centers, kpar_centers, f[this_band].item()['sample_2D_mK2']**2 +  f[this_band].item()['thermal_2D_mK2']**2

def results_fig4(kperp,kpar, noisy_sample, mean, pred, redshift, 
                 cmap='viridis', cmap2='cividis', fname=None, 
                 vminfe=None, vmaxfe=None, fontsize=30, figsize=(18,16), 
                 label3 = r'$\frac{|\Delta^2_{21} - \Delta^2_{21,test}|}{\sigma_{\rm HERA}}$'):
    r"""Plot Figure 4 in Breitman+25a."""
    rcParams.update({'font.size': fontsize})
    fig, ax = plt.subplots(nrows=3, ncols=2,sharex=True, sharey=True, figsize = figsize, layout='constrained')
    vmin = np.percentile(mean, 2.5)
    vmax = np.percentile(mean,97.5)
    ax[0,0].pcolormesh(kperp, kpar,noisy_sample.T, vmin = vmin, vmax = vmax)
    ax[0,0].text(0.25,0.7,'Sample', color = 'k')
    ax[0,0].text(0.25,0.42,'z ~ ' + str(np.round(redshift,1)), color = 'k')

    im = ax[0,1].pcolormesh(kperp, kpar, pred.T, vmin = vmin, vmax = vmax)
#     ax[0,1].text(0.5,0.7,'Mean', color = 'k')
#     im = ax[0,2].pcolormesh(kperp, kpar, pred.T, vmin = vmin, vmax = vmax)
    ax[0,1].text(0.043,0.7,'py21cmDenoiser', color = 'k')
    fig.colorbar(im, label=r'$\Delta^2_{21}$ [mK$^2$]')
    mu_fe = abs((noisy_sample - mean)/mean)*100.#, axis = (-1,-2))
    #abs_err = abs(pred - mean)
    if vminfe is None:
        vminfe = np.nanpercentile(mu_fe, 2.5)
    if vmaxfe is None:
        vmaxfe = np.nanpercentile(mu_fe,97.5)
    ax[1,0].pcolormesh(kperp, kpar,mu_fe.T, norm=colors.LogNorm(vmin=vminfe, vmax=vmaxfe),cmap = cmap)
    mu_fe = abs((pred - mean)/mean)*100.
    im = ax[1,1].pcolormesh(kperp, kpar,mu_fe.T, norm=colors.LogNorm(vmin=vminfe, vmax=vmaxfe),cmap = cmap)
    fig.colorbar(im, label=r'FE(%)')
    
#     im = ax[1,1].pcolormesh(kperp, kpar,abs_err.T, cmap = 'YlGnBu')
#     plt.colorbar(im, label='Abs Diff [mK$^2$]')
    # Mock
    hera_z, hera_kperp, hera_kpar, hera_var = get_hera_var_at_z(redshift)
    hera_var[np.isinf(hera_var)] = np.nan
    interped_pred = RectBivariateSpline(kperp, kpar, pred)(hera_kperp,hera_kpar)
    interped_inp = RectBivariateSpline(kperp, kpar, noisy_sample)(hera_kperp,hera_kpar)
    interped_mean = RectBivariateSpline(kperp, kpar, mean)(hera_kperp,hera_kpar)
    #interp pred and mean to hera grid
    hera_err = abs((interped_pred - interped_mean)/np.sqrt(hera_var))
    hera_err_noisy = abs((interped_inp - interped_mean)/np.sqrt(hera_var))
    vmin = 1e-2
    vmax = 1.
    ax[2,0].pcolormesh(kperp, kpar, hera_err_noisy.T, 
                       norm=colors.LogNorm(vmin=vmin, vmax=vmax),
                       cmap = cmap2)
    im = ax[2,1].pcolormesh(kperp, kpar,hera_err.T, 
                            norm=colors.LogNorm(vmin=vmin, vmax=vmax),
                            cmap = cmap2)
    cbar = fig.colorbar(im, ticks = [1, 0.1, 0.01, 0.001],
                 label=label3)
    
    min_hera_kperp = 9e-3 #hMpc^{-1}
    try:
        kperp = kperp.value
    except:
        pass
    try:
        hlb = hlb.value
    except:
        pass
    hera_c = 'b'
    
    kperpp = np.linspace(4e-3, 3.,1000)/u.Mpc
    hl = horizon_limit(kperpp, z2freq(redshift)*u.MHz, buffer = 0*u.ns).value
    hlb = horizon_limit(kperpp, z2freq(redshift)*u.MHz, buffer = 300*u.ns).value

    m = kperpp.value >= min_hera_kperp * cosmo.h
    ax[2,0].plot(kperpp[m], hlb[m], color = 'k', lw=2, label = 'HERA 300 ns buffer')
    ax[2,1].plot(kperpp[m], hlb[m], color = 'k', lw=2)

    ax[2,0].fill_between(kperpp[m].value, np.zeros(len(hlb[m])),hlb[m], color = 'k', alpha = 0.3, zorder=1)
    ax[2,1].fill_between(kperpp[m].value, np.zeros(len(hlb[m])),hlb[m], color = 'k', alpha = 0.3, zorder=1)

    ax[2,0].vlines(kperpp[m][0].value, hlb[m][0], hlb[-1], lw = 2, color = 'k')
    ax[2,1].vlines(kperpp[m][0].value, hlb[m][0], hlb[-1], lw = 2, color = 'k')
    #ax[1,2].set_xlim(kperpp[m][0].value/cosmo.h,kperp[-1]/cosmo.h)
    ff = interp1d(kperpp.value, hl, bounds_error=False)
    xx = np.logspace(np.log10(kperpp[0].value), np.log10(kperp[-1]))
    ax[2,0].plot(xx, ff(xx), color = 'cyan', ls = '--', zorder = 23,lw = 2, alpha = 0.8, label = 'Horizon limit')
    ax[2,1].plot(xx, ff(xx), color = 'cyan', ls = '--', zorder = 23,lw = 2, alpha = 0.8)

    xvals = np.linspace(kperpp[0].value, kperpp[-1].value, 100)
    mus = [0.97]
    cs = ['r', 'orange']
    r = np.logspace(-10,1., 10000)
    for ii, muval in enumerate(mus):
        theta = np.arccos(muval)
        x = r*np.sin(theta)
        y = r*muval
        m = x > kperpp[0].value
        ax[2,0].plot(x[m],y[m],color = cs[ii], lw = 2, zorder = 23,label = r'$\mu_{\rm min}$ = ' + str(muval))
        ax[2,1].plot(x[m],y[m],color = cs[ii], lw = 2, zorder = 23)

    ax[2,0].set_xlabel(r'k$_\perp$ [Mpc$^{-1}$]')
    ax[2,1].set_xlabel(r'k$_\perp$ [Mpc$^{-1}$]')
    ax[2,0].set_ylabel(r'k$_\parallel$ [Mpc$^{-1}$]')
    ax[0,0].set_ylabel(r'k$_\parallel$ [Mpc$^{-1}$]')
    ax[1,0].set_ylabel(r'k$_\parallel$ [Mpc$^{-1}$]')
    ax[2,0].legend(frameon=False, loc = 4, fontsize = np.min([fontsize, 18]))

    for a in ax.ravel():
        a.set_xlim(kperp[0], kperp[-1])
        a.set_ylim(kpar[0], kpar[-1])
    plt.loglog()
    if fname is not None:
        plt.savefig(fname)
    plt.show()
    
def results_fig5(kperp, kpar,all_mu_fes, all_sv_fe, all_zs, cmap = 'cividis', z_range=[8,9], qt = '50', path=None, fontsize = 30, figsize=(18,7)):
    r"""Plot Figure 5 in Breitman+25a."""
    rcParams.update({'font.size': fontsize})
    qt = str(qt)
    m = np.logical_and(all_zs >= z_range[0], all_zs<=z_range[1])
    print("There are",sum(m),"samples in the redshift range requested.")
    fe_percs_sv = np.percentile((all_sv_fe)[m], [2.5,16,50,84,97.5], axis = 0)
    fe_percs = np.percentile((all_mu_fes)[m], [2.5,16,50,84,97.5], axis = 0)

    #abs_err_percs = np.percentile((all_sv_abserr)[m], [2.5,16,50,84,97.5], axis = 0)

    fig, ax = plt.subplots(nrows = 1, ncols = 2,sharex=True, sharey=True, figsize = figsize, layout='constrained')
    ax = np.atleast_2d(ax)
    if qt == '50':
        qt_sv = fe_percs_sv[2]
        qt_ml = fe_percs[2]
        label = 'Median '
    elif qt == '68':
        qt_sv = (fe_percs_sv[3] - fe_percs_sv[1])
        qt_ml = (fe_percs[3] - fe_percs[1])
        label = r'1$\sigma$ '
    elif qt == '95':
        qt_sv = (fe_percs_sv[4] - fe_percs_sv[0])
        qt_ml = (fe_percs[4] - fe_percs[0])
        label = r'2$\sigma$ '
    else:
        raise ValueError('qt is 50, 68 or 95.')
    vmin = np.min([0.8,np.nanpercentile(qt_sv, 2.5)])
    vmax = np.nanpercentile(qt_sv, 97.5)
    im = ax[0,1].pcolormesh(kperp, kpar, qt_ml.T, norm = colors.LogNorm(vmin=vmin, vmax=vmax) if vmin > 0 else colors.SymLogNorm(vmin=vmin, vmax=vmax, linthresh=0.1), cmap = cmap,)
    plt.colorbar(im, label=label+'FE (%)')
    ax[0,0].pcolormesh(kperp, kpar, qt_sv.T,norm = colors.LogNorm(vmin=vmin, vmax=vmax) if vmin > 0 else colors.SymLogNorm(vmin=vmin, vmax=vmax, linthresh=0.1),  cmap = cmap,)

#     vmin = np.nanpercentile(fe_percs_sv[3] - fe_percs_sv[1], 2.5)
#     vmax = np.nanpercentile(fe_percs_sv[3] - fe_percs_sv[1], 97.5)

#     im = ax[1,1].pcolormesh(kperp, kpar, (fe_percs[3] - fe_percs[1]).T, norm = colors.LogNorm(vmin=vmin, vmax=vmax), cmap = cmap,)
#     plt.colorbar(im, label='FE(%) 68% CL')

#     ax[1,0].pcolormesh(kperp, kpar, (fe_percs_sv[3] - fe_percs_sv[1]).T, norm = colors.LogNorm(vmin=vmin, vmax=vmax), cmap = cmap,)

    plt.loglog()    
    
    ax[0,0].text(0.2,0.7,'Sample', color = 'white')
    ax[0,0].text(0.2,0.42,'z ~ ' + str(np.round(np.mean(all_zs[m]),1)), color = 'white')
    ax[0,1].text(0.03,0.7,'py21cmDenoiser', color = 'white')
    
    ax[0,0].set_xlabel(r'k$_\perp$ [Mpc$^{-1}$]')
    ax[0,1].set_xlabel(r'k$_\perp$ [Mpc$^{-1}$]')
    ax[0,0].set_ylabel(r'k$_\parallel$ [Mpc$^{-1}$]')
    #ax[1,0].set_ylabel(r'k$_\parallel$ [Mpc$^{-1}$]')
    if path is not None:
        plt.savefig(path+'NN_'+qt+'_overtest_z'+str(int(np.round(np.mean(all_zs[m]))))+'_sv.png')
    plt.show()